/**
 * A helper that checks if the given argument is of a certain type.
 */
export function $is(type: BigIntConstructor, arg: any): arg is bigint;
export function $is(type: BooleanConstructor, arg: any): arg is boolean;
export function $is(type: NumberConstructor, arg: any): arg is number;
export function $is(type: StringConstructor, arg: any): arg is string;
export function $is(type: SymbolConstructor, arg: any): arg is symbol;
export function $is(type: undefined, arg: any): arg is null | undefined;
export function $is<T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;

/**
 * A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export function $in(key: any, ref: any): boolean;

/**
 * A helper that checks equality of the given arguments.
 */
export function $eq(x: any, y: any): boolean;

/**
 * The type of arguments of the `$h` and `$s` helpers.
 */
export type $HArgs = [
  string | Node,
  ...(Record<string, any> | null | undefined | Node | string | number | $HArgs)[]
];

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `$s`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttribute`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `$HArgs` are passed to `$h` and the results are appended to the element being created or modified.
 */
export function $h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args: $HArgs[1][]): HTMLElementTagNameMap[T];
export function $h<N extends Node>(node: N, ...args: $HArgs[1][]): N;
export function $h(...args: $HArgs): Node;

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `$h`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttributeNS`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttributeNS`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `$HArgs` are passed to `$s` and the results are appended to the element being created or modified.
 */
export function $s<T extends keyof SVGElementTagNameMap>(tag: T, ...args: $HArgs[1][]): SVGElementTagNameMap[T];
export function $s<N extends Node>(node: N, ...args: $HArgs[1][]): N;
export function $s(...args: $HArgs): Node;

/**
 * The type of arguments of the `$c` helper.
 */
export type $CNode = {
  [attributeOrSelector: string]: string | number | $CNode;
};

/**
 * The type of arguments of the `$c` helper.
 */
export type $CRoot = Record<string, $CNode>;

/**
 * A simple CSS-in-JS helper.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export function $c(root: $CRoot, splitter?: string): string;

/**
 * A generic helper for syntax highlighting (see also `$highlightJs`).
 */
export function $highlight(
  pattern: RegExp,
  highlighters: ((chunk: string, index: number) => $HArgs[1])[],
  codeJs: string
): $HArgs[1][];

/**
 * A helper for highlighting JavaScript.
 */
export function $highlightJs (codeJs: string): $HArgs[1][];

/**
 * A helper that implements typographic corrections specific to Polish typography.
 */
export function $fixTypography(node: Node): void;

/**
 * Language translations helper.
 */
export function $locale(
  locales: Record<string, Record<string, string | Record<string, string>>>,
  defaultLanguage: string,
  languages?: readonly string[]
): (text?: string, version?: string) => string;

/**
 * A convenient helper for getting values of nested objects.
 */
export function $get(ref: any, keys: (string | number | symbol)[], defaultValue?: any): any;

/**
 * A convenient helper for setting values of nested objects.
 */
export function $set(ref: any, keys: (string | number | symbol)[], value: any): void;

/**
 * A helper for choosing the correct singular and plural.
 */
export function $plUral(singular: string, plural2: string, plural5: string, value: number): string;

/**
 * A helper that generates a UUID v1 identifier (with a creation timestamp).
 *
 * - The optional `node` parameter should have the format `/^[0123456789abcdef]+$/`.
 *   Its value will be trimmed to last 12 characters and left padded with zeros.
 */
export function $uuid1(options?: {
  date?: Date;
  node?: string;
}): string;

/**
 * The type of arguments of the `$sql` helper.
 */
export type $EscapeMap = Map<any, (value?: any) => string>;

/**
 * A generic helper for escaping SQL queries.
 */
export function $sql(escapeMap: $EscapeMap): (template: TemplateStringsArray, ...substitutions: any[]) => string;
